'''
    guide-search
    -----------
    This module provides the interface to elasticsearch
    :copyright: (c) 2017 John Pickerill.
    :license: MIT/X11, see LICENSE for more details.
'''

# -*- coding: utf-8 -*-

from .__about__ import __version__, __title__
from .esearch import Esearch
from .dsl import Dsl
from .exceptions import (    # noqa: F401
    BadRequestError,            # 400
    ResourceNotFoundError,      # 404
    ConflictError,              # 409
    PreconditionError,          # 412
    ServerError,                # 500
    ServiceUnreachableError,    # 503
    UnknownError,               # unexpected htstp response
    ResultParseError,           # es result not in expected form
    CommitError,
    JSONDecodeError,
    ValidationError)

__service_name__ = __title__

__all__ = [ 'Esearch',
            'Dsl',
            'ResourceNotFoundError',
            '__version__']



from collections import OrderedDict


class Dsl(object):
    buffer = {}
    query = {}
    es1 = False

    def __init__(self):
        self.buffer = {}

    def dsl(self):
        return self.buffer

    def ui_query(self, search):
        # based on the search structure
        # create a filter to determine what is in the search results list
        # and a query to determine their order

        pre_filter = search.get("filter", [])

        if search['free_text']:
            and_filter = [{"multi_match": {"query": search['free_text'], "type": "cross_fields", "fields": search['fields'],
                                       "operator": "and"}}]
            shoulds = [{"multi_match": {"query": search['free_text'],"type": "most_fields", "fields": search['fields']}}]
        else:
            and_filter = []
            shoulds = []

        for phrase in search['exact_phrases']:
            exact_filter = {"multi_match": {"query": phrase, "type": "phrase", "fields": search['fields']}}
            and_filter.append(exact_filter)
            shoulds.append(exact_filter)

        pre_filter.append({'bool': {'must': and_filter}})
        if "facets" in search:
            if len(search['facets']):
                pre_filter.append(self.m_facet_filter(search["facets"]))

        query = {
            "query": {
                "bool": {
                     "should": shoulds
                }
            }
        }

        if len(pre_filter):
            query["query"]["bool"]["filter"] = pre_filter
        self.buffer.update(query)
        return self


    def Q(self):
        self.query = {"bool": {}}
        self.buffer.update({'query': self.query})
        return self

    def match_all(self, filter=None):
        if filter:
            self.query = {"bool": {"must": [{"match_all": {}}], "filter": filter}}
        else:
            self.query = {"bool": {"must": [{"match_all": {}}]}}
        self.buffer.update({'query': self.query})
        return self

    def Q_must(self, must):
        if "must" not in self.query["bool"]:
            self.query["bool"]["must"] = []
        self.query["bool"]["must"].append(must)
        return self

    def Q_should(self, should):
        if "should" not in self.query["bool"]:
            self.query["bool"]["should"] = []
        self.query["bool"]["should"].append(should)
        return self

    def Q_filter(self, filter):
        if "filter" not in self.query["bool"]:
            self.query["bool"]["filter"] = {'bool': {'must': []}}
        self.query["bool"]["filter"]['bool']['must'].append(filter)


    def more_like_this(self, index, source):
        if self.es1:
            buffer = {"query": {"bool": {"should": source}}, "_source": ["id", "title", "scope"]}
        else:
            buffer = {
                    "_source": ["id", "title", "scope"],
                    "query": {
                        "more_like_this": {
                            "fields": ["keywords", "scope"],
                            "like": [{
                                "_index": index,
                                "_type": "articles",
                                "_id": source}],
                            "min_term_freq": 1,
                            "max_query_terms": 12
                            }}}
        self.buffer.update(buffer)
        return self

    def item_use(self, id, size=1000):
        if self.es1:
            buffer = {"fields": ["id"], "query": {"term": {"items.item": id}}, "size": 1000}  # es1.4
            self.buffer.update(buffer)
        else:
            self.match_all({"term": {"items.item": id}}).fields('id').size(size)  # es5.2
        return self

    def associate_list(self, aType):
        buffer = {"query": {"term": {"type": aType}},
                  "_source": ["id", "title"]}
        self.buffer.update(buffer)
        return self

    def landing_list(self):
        buffer = {"_source": ["id", "title", "scope"],
                  "query": {"term": {"purpose": "landing"}},
                  "sort": {"id": "asc"}}
        self.buffer.update(buffer)
        return self

    def cluster_list(self, max=10000):
        buffer = {"size": "0",
                  "aggs": {
                             "clusters": {
                                "nested": {"path": "clusters"},
                                "aggs":  {"clusters":
                                          {"terms":
                                           {"field": "clusters.cluster", "order": {"_term": "asc"}, "size": max},
                                           "aggs": {
                                                        "sens": {
                                                            "reverse_nested": {},
                                                            "aggs": {"sens": {"terms": {"field": "sensitivity"}}}
                                                        }
                                                    }
                                           },
                                          "cluster_count": {"cardinality": {"field": "clusters.cluster"}}}
                                }}}
        self.buffer.update(buffer)
        return self

    def query_cluster(self, id):
        if self.es1:
            buffer = {"query": {"nested": {"path": "clusters", "filter": {"term": {"cluster": id}}}},
                      "fields": ["id", "title", "scope", "sensitivity"],
                      "sort": [{"clusters.priority": {
                                "order": "asc",
                                "nested_path": "clusters",
                                "nested_filter": {"term": {"clusters.cluster": id}}}}]}
        else:
            buffer = {"query": {"nested": {
                                    "path": "clusters",
                                    "query": {"bool": {"filter": {"term": {"clusters.cluster": id}}}}}},
                      "_source": ["id", "title", "scope", "sensitivity"],
                      "sort": [{"clusters.priority": {
                                "order": "asc",
                                "nested_path": "clusters",
                                "nested_filter": {"term": {"clusters.cluster": id}}}}]}
        self.buffer.update(buffer)
        return self

    def fields(self, fields):
        self.buffer.update({'_source': fields})
        return self

    def size(self, size, page=0):
        ctrl = {"size": int(size)}
        page = int(page)
        if page > 0:
            page = page - 1
            if page > 0:
                ctrl["from"] = page * ctrl["size"]
        self.buffer.update(ctrl)
        return self

    def sort(self, sort="_score", order="desc"):
        # TODO this is a non-generic bit for UI need to think about factoring this out to somewhere else
        self.buffer.update({"sort":  [{sort: {"order": order}}]})
        return self

    def suggest(self, data, source, max=10):
        self.buffer.update({"suggest": {"didyoumean": {"text": data, "term": {"field": source, "size": max}}}})
        return self

    def facets(self, max=50):
        self.buffer.update({"aggs": {
                              "nest":
                              {"nested":
                               {"path": "facets"},
                               "aggs": {"facetnames":
                                        {"terms":
                                         {"field": "facets.facet"},
                                         "aggs": {"focinames":
                                                  {"terms": {"field": "facets.foci", "size": max}}
                                                  }}}}}})
        return self

    def m_facet_filter(self, fociSelected):
        selected = OrderedDict()
        for items in fociSelected:
            facet_name, foci_name = items.split('|')
            if facet_name not in selected:
                selected[facet_name] = []
            selected[facet_name].append(foci_name)
        filters = []
        for fa, foci in selected.items():
            musts = [{"term": {"facets.facet": fa}}]

            for fo in foci:
                musts.append({"terms": {"facets.foci": [fo]}})
            if self.es1:
                facet = {"nested": {"path": "facets", "filter": {"bool": {"must": musts}}}}
            else:
                facet = {"nested": {"path": "facets", "query": {"bool": {"must": musts}}}}
            filters.append(facet)

        if self.es1:
            filter = {"and": {"filters": filters}}
        else:
            filter = {"bool": {"filter": filters}}
        return filter



import os
import pytest
import requests
from urllib.parse import splitport
from guide_search.esearch import Esearch, Dsl
from utils import log, AttrDict, Helper
from guide_search import exceptions

DEBUG=True
GUIDE_API_HOST=os.environ['GUIDE_API_HOST']
ES_HOST=os.environ['ES_HOST']

@pytest.fixture
def esearch(setup):
    return setup.esearch


@pytest.fixture
def index_name(setup):
    return setup.index_name


@pytest.fixture(scope="module")
def setup():
    """
    - config is created once for all tests
    - initialises index
    """
    setup = AttrDict() # {{{

    setup.esearch = Esearch(
            host=ES_HOST,
            control='knowledge_config/control' )


    setup.helper = Helper(GUIDE_API_HOST)

    if not DEBUG:

        # delete any test indexes
        stats = requests.get(setup.esearch.host+'/_stats').json()
        indices = stats['indices']
        test_indices = [ i for i in list(indices.keys()) if i.startswith('test_') ]
        for i in test_indices:
            log('Deleting index %s' % i)
            r=requests.delete(setup.esearch.host+'/'+i).json()
            if not r or not r['acknowledged']:
                log('Failed to deleted index %s' % i)

        # recreate branch from last commit on master and update ES
        setup.index_name = setup.helper.randomName() # random branch name e.g. test_imjnxy
        setup.helper.delete_test_branches()
        setup.helper.initialise_index(setup.index_name)
    else:
        # re-use branch / ES content
        setup.index_name = setup.helper.get_test_branch_names()[0]

    yield setup
# }}}


def test_check_versions(setup):
    """ guide-api is running and connected to the right version servers """
    log('')
    host = setup.helper.info['elasticsearch']['host']
    ver = setup.helper.info['elasticsearch']['version']
    version = [int(n) for n in ver.split('.')]
    assert version[0] >= 5 and \
        log('...detected Elasticsearch version %s on %s' % (ver, host))

    host = setup.helper.info['gitlab']['content']['host']
    ver = setup.helper.info['gitlab']['content']['version']
    version = [int(n) for n in ver.split('.')]
    assert version[0] >= 8 and \
        log('...detected Gitlab version %s on %s' % (ver, host))


def test_request_base_url(esearch):
    esearch.Requests=requests
    url = esearch.make_url('')
    log('getting %s ...' % url)
    r = esearch.request(method='get', url=url)
    expected = 'version' in r.keys() 
    assert expected and log('Base ES url (%s) returns version info ' % url)


def test_request_bad_url(esearch):
    url = esearch.make_url("")
    url,port = splitport(url)
    url = '{}:{}'.format(url,int(port)+1)
    with pytest.raises(requests.exceptions.ConnectionError):
        esearch.request(method='get', url=url)
    log('Bad url (%s) raises a ConnectionError exception ' % url)


def test_request_bad_route(esearch, index_name):
    with pytest.raises(exceptions.BadRequestError):
        url = esearch.make_url(index_name, "spaghetti")
        esearch.request(method='get', url=url, params=None, data=None, raw_response=False)
    log('Right url + bad route (%s) raises a BadRequestError exception ' % url)


# def test_search_match_all_articles(esearch, index_name):
    # from pudb import set_trace; set_trace()
    # hits = esearch.search(index=index_name, type='articles', dsl= Dsl().match_all().buffer)
    # n=len(hits)
    # assert n and log('match all returns %s articles' % n)




# curl 'elasticsearch5:9200/test_g4n6yb/_search?q=mermaid'
# @pytest.mark.parametrize('id', ['acc_fence'])
# def test_search(esearch, index_name, id):
    # """
    # search successfully queries elasticsearch for a document by its ID.

    # https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-ids-query.html#query-dsl-ids-query

    # """
    # dsl = {
        # 'query': {
            # 'ids': { 
            # 'values' : [id],
            # }
        # },
        # 'size': 2,
    # }
    # r = esearch.search(index=index_name, type="articles", dsl=dsl)
    # assert len(r['hits']['hits']) == 1
    # assert r['hits']['total'] == 1
    # assert r['hits']['hits'][0]['_id'] == id




__title__ = 'guide-search'
__description__ = 'Guidance management api'
__url__ = 'https://github.com/johnpickerill/guide-search'
__version_info__ = ('0', '0', '66')
__version__ = '.'.join(__version_info__)
__author__ = 'John Pickerill'
__author_email__ = 'john.pickerill@landregistry.gov.uk'
__maintainer__ = 'Land Registry'
__license__ = 'MIT'
__copyright__ = '(c) 2017 by John Pickerill'



# vim: foldmethod=marker:foldmarker=\ {{{,\ }}}:cms=#%s
import requests
import os
from random import Random
from termcolor import cprint

def log(s):
    cprint('%s' % s, 'cyan')
    return True


class AttrDict(dict):
    """ dictionary that exposes its keys as attributes """
    def __init__(self, *args, **kwargs):  # {{{
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self # }}}

class Helper(object):
# {{{
    RANDCHARS = [chr(i) for i in range(97,123)]
    RANDCHARS.extend( [chr(i) for i in range(48,58)] )

    def randomName(self, prefix='test_', n=6):
        r = Random()
        prefix = 'test_'
        for i in range(6):
            prefix = prefix + self.RANDCHARS[r.randint(0, len(self.RANDCHARS)-1)]
        return prefix

    def __init__(self, guide_api_host):
        # get the gitlab server details from guide-api
        self.api_host = guide_api_host
        self.info = requests.get(self.api_host+'/info').json() #  'http://hh-kng-d11.lnx.lr.net'
        self.gitlab_host = self.info['gitlab']['content']['host']
        self.project_id = self.info['gitlab']['content']['project'] # 13
        self.headers = {'PRIVATE-TOKEN': '45PpX1Y-n2Ygi6EyyuZ7'} # TODO: move into config or add a create branch endpoint to guide-api

    def get_commits(self, branch_name='master'):

        """ https://docs.gitlab.com/ee/api/branches.html#create-repository-branch  """

        url = '%s/api/v3/projects/%s/repository/commits' % (self.gitlab_host, self.project_id)
        params = {'ref_name': branch_name }
        kwargs = { 'headers':self.headers, 'params': params }
        return requests.get(url, **kwargs)


    def create_branch(self, branch_name, sha1='empty'):

        """ https://docs.gitlab.com/ee/api/branches.html#create-repository-branch """

        url = '%s/api/v3/projects/%s/repository/branches' % (self.gitlab_host, self.project_id)
        params = {'branch_name': branch_name, 'ref': sha1}
        kwargs = { 'headers':self.headers, 'params': params }
        return requests.post(url, **kwargs)


    def delete_branch(self, branch_name):

        """ https://docs.gitlab.com/ee/api/branches.html#delete-repository-branch """

        url = '%s/api/v3/projects/%s/repository/branches/%s' % (self.gitlab_host, self.project_id, branch_name)
        kwargs = { 'headers':self.headers, 'params': {} }
        return requests.delete(url, **kwargs)
    
    def list_branches(self):

        """ https://docs.gitlab.com/ee/api/branches.html#list-repository-branches """

        url = '%s/api/v3/projects/%s/repository/branches' % (self.gitlab_host, self.project_id)
        kwargs = { 'headers':self.headers, 'params': {} }
        return requests.get(url, **kwargs)


    def delete_test_branches(self):

        """ Delete all branches prefixed with 'test_' """

        for test_branch in self.get_test_branch_names():
            log('Deleting branch %s..' % test_branch)
            r = self.delete_branch(test_branch)
            if not r or r.status_code != 200:
                log('Failed to delete branch %s..' % test_branch)

    def get_test_branch_names(self):
        r = self.list_branches()
        names = [b['name'] for b in r.json()]
        return [n for n in names if n.startswith('test_')]


    def initialise_index(self, index_name):

        """ Delete all branches prefixed with 'test_' """

        Host = os.environ['GUIDE_API_HOST']
        log('HOST=%s' % host)

        # curl -X PUT http://localhost:5002/api/initialise/controls
        # upload files from git repo: dev-guidance / configuration to ES5
        url = host+'/api/initialise/controls'
        log('\nInitialising controls %s ...' % url)
        r = requests.put(url)
        for k,v in r.json().items():
            log("\t%s = %s" % (k,v))
            assert v == 'success'

        # curl -X PUT http://localhost:5002/api/initialise/index/test_a3J2/test_a3J2
        # put more stuff in ES
        url = host+'/api/initialise/index/'+index_name+'/'+index_name
        log('\nInitialising index %s ...' % url)
        r = requests.put(url)
        assert(r.json()['acknowledged'])

        # get last commit on master branch
        commits = self.get_commits(branch_name='master').json()
        sha1 = commits[0]['id']

        # create a branch of the last commit
        r = self.create_branch(index_name, sha1)

        assert r.status_code == 201
        commit=r.json()
        assert commit

        
        # curl -X PUT http://localhost:5002/api/publish/test_a3J2/_publish
        # publish the content in the above branch to ES
        url = host+'/api/publish/'+index_name+'/_publish'
        log('\nPublishing content %s ...' % url)
        r = requests.put(url)
        assert r.json()['status']=='started'

# }}}

